from typing import Dict, Optional
from functools import wraps
import threading

from libbs.api import DecompilerInterface


class AIAPI:
    def __init__(
        self,
        decompiler_interface: Optional[DecompilerInterface] = None,
        decompiler_name: Optional[str] = None,
        use_decompiler: bool = True,
        delay_init: bool = False,
        # size in bytes
        min_func_size: int = 0x10,
        max_func_size: int = 0xffff,
        model=None,
    ):
        # useful for initing after the creation of a decompiler interface
        self._dec_interface: DecompilerInterface = None
        self._dec_name = None
        self._delay_init = delay_init
        if not self._delay_init:
            self.init_decompiler_interface(decompiler_interface, decompiler_name, use_decompiler)

        self._min_func_size = min_func_size
        self._max_func_size = max_func_size
        self.model = model or self.__class__.__name__

    def init_decompiler_interface(
        self,
        decompiler_interface: Optional[DecompilerInterface] = None,
        decompiler_name: Optional[str] = None,
        use_decompiler: bool = True
    ):
        self._dec_interface: DecompilerInterface = DecompilerInterface.discover(force_decompiler=decompiler_name) \
            if use_decompiler and decompiler_interface is None else decompiler_interface
        self._dec_name = decompiler_name if decompiler_interface is None else decompiler_interface.name
        if self._dec_interface is None and not self._dec_name:
            raise ValueError("You must either provide a decompiler name or a decompiler interface.")

    def info(self, msg):
        if self._dec_interface is not None:
            self._dec_interface.info(msg)

    def debug(self, msg):
        if self._dec_interface is not None:
            self._dec_interface.debug(msg)

    def warning(self, msg):
        if self._dec_interface is not None:
            self._dec_interface.warning(msg)

    def error(self, msg):
        if self._dec_interface is not None:
            self._dec_interface.error(msg)

    @property
    def has_decompiler_gui(self):
        return self._dec_interface is not None and not self._dec_interface.headless

    @staticmethod
    def requires_function(f):
        """
        A wrapper function to make sure an API call has decompilation text to operate on and possibly a Function
        object. There are really two modes any API call operates in:
        1. Without Decompiler Backend: requires provided dec text
        2. With Decompiler Backend:
               2a. With UI: Function will be collected from the UI if not provided
               2b. Without UI: requires a FunctionA

        The Function collected from the UI is the one the use is currently looking at.
        """
        @wraps(f)
        def _requires_function(*args, ai_api: "AIAPI" = None, **kwargs):
            function = kwargs.pop("function", None)
            dec_text = kwargs.pop("dec_text", None)
            use_dec = kwargs.pop("use_dec", True)
            has_self = kwargs.pop("has_self", True)
            # make the self object the new AI API, should only be used inside an AIAPI class
            if not ai_api and has_self:
                ai_api = args[0]

            if not dec_text and not use_dec:
                raise ValueError("You must provide decompile text if you are not using a dec backend")

            # two mode constructions: with decompiler and without
            # with decompiler backend
            if use_dec:
                if not ai_api.has_decompiler_gui and function is None:
                    raise ValueError("You must provide a Function when using this with a decompiler")

                # we must have a UI if we have no func
                if function is None:
                    function = ai_api._dec_interface.functions[ai_api._dec_interface.gui_active_context().addr]

                # get new text with the function that is present
                if dec_text is None:
                    dec_text = ai_api._dec_interface.decompile(function.addr)

            return f(*args, function=function, dec_text=dec_text, use_dec=use_dec, **kwargs)

        return _requires_function




# Run the DAILA server and open the DAILA selector UI.
# @author mahaloz
# @category AI
# @menupath Tools.DAILA.Start DAILA Backend


python_library_command = "dailalib -s ghidra"
shell_library_command = "daila -s ghidra"
def create_plugin(*args, **kwargs):
    from dailalib import create_plugin as _create_plugin
    return _create_plugin(*args, **kwargs)

# =============================================================================
# LibBS generic plugin loader (don't touch)
# =============================================================================

import sys
# Python 2 has special requirements for Ghidra, which forces us to use a different entry point
# and scope for defining plugin entry points
if sys.version[0] == "2":
    # Do Ghidra Py2 entry point
    import subprocess
    from libbs_vendored.ghidra_bridge_server import GhidraBridgeServer
    from distutils.spawn import find_executable
    cmd = shell_library_command.split(" ") if shell_library_command else python_library_command.split(" ")
    if not find_executable(cmd[0]):
        # fallback to doing a python command
        cmd = python_library_command.split(" ")

    GhidraBridgeServer.run_server(background=True)
    process = subprocess.Popen(cmd)
    if process.poll() is not None:
        raise RuntimeError(
            "Failed to run the Python3 backed. It's likely Python3 is not in your Path inside Ghidra.")
else:
    # Try plugin discovery for other decompilers
    try:
        import idaapi
        has_ida = True
    except ImportError:
        has_ida = False
    try:
        import angrmanagement
        has_angr = True
    except ImportError:
        has_angr = False

    if not has_ida and not has_angr:
        create_plugin()
    elif has_angr:
        from angrmanagement.plugins import BasePlugin
        class AngrBSPluginThunk(BasePlugin):
            def __init__(self, workspace):
                super().__init__(workspace)
                globals()["workspace"] = workspace
                self.plugin = create_plugin()

            def teardown(self):
                pass


def PLUGIN_ENTRY(*args, **kwargs):
    """
    This is the entry point for IDA to load the plugin.
    """
    return create_plugin(*args, **kwargs)



import logging

from binsync.data import Function, State

from dailalib.binsync_plugin.ai_bs_user import AIBSUser

_l = logging.getLogger(__name__)


class VARModelBSUser(AIBSUser):
    DEFAULT_USERNAME = "varmodel_user"

    """
    Variable Annotation Recovery (VAR) Model
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            from varmodel import VariableRenamingAPI
        except ImportError:
            _l.error("VARModel is not installed and is still closed source. You will be unable to use this BinSync user.")
            return

        self._renaming_api = VariableRenamingAPI()
    
    def run_all_ai_commands_for_dec(self, decompilation: str, func: Function, state: State):
        try:
            updated_func: Function = self._renaming_api.predict_variable_names(decompilation, func)
        except Exception as e:
            _l.warning(f"Skipping {func} due to exception: {e}")
            return 0

        if updated_func is not None and (updated_func.args or updated_func.stack_vars):
            # count changes
            changes = len(updated_func.args) + len(updated_func.stack_vars)
            state.set_function(updated_func)
            return changes

        return 0
